from avatar_utils.validation.schemas import SuccessResponseSchema, ErrorResponseSchema
from avatar_utils.validation.validate_json import validate_json

__all__ = ['validate_json', 'SuccessResponseSchema', 'ErrorResponseSchema']



from logging import getLogger

from marshmallow import Schema, fields
from marshmallow.validate import Equal
from werkzeug.exceptions import BadRequest  # noqa: PyPackageRequirements

__all__ = ['SuccessResponseSchema', 'ErrorResponseSchema']

logger = getLogger('sample-service')


class ResponseSchema(Schema):
    headers = fields.Dict(required=True, allow_none=False, default=dict())
    message = fields.Str(required=True, allow_none=True, default='')
    result = fields.Dict(required=True, allow_none=True)
    success = fields.Bool(required=True, allow_none=False)


class SuccessResponseSchema(ResponseSchema):
    success = fields.Bool(required=True, allow_none=False, validate=Equal(True), default=True)


class ErrorResponseSchema(ResponseSchema):
    success = fields.Bool(required=True, allow_none=False, validate=Equal(False), default=False)



import traceback
from functools import wraps
from logging import getLogger
from typing import Type, Optional, Union, Callable, Tuple

from flask import request, jsonify
from marshmallow import Schema
from marshmallow import ValidationError
from werkzeug import Response, exceptions  # noqa: PyPackageRequirements

from avatar_utils.core import create_response
from avatar_utils.validation.constants import ServiceHTTPMethods, ServiceHTTPCodes

logger = getLogger(__name__)


def validate_json(request_schema: Optional[Type[Schema]] = None,
                  response_schema: Optional[Type[Schema]] = None,
                  method: Union[str, ServiceHTTPMethods] = ServiceHTTPMethods.GET) -> Callable:

    method = method.name if isinstance(method, ServiceHTTPMethods) else method.upper()

    def decorator(f) -> Callable:

        @wraps(f)
        def wrapper(*args, **kw) -> Union[str, Tuple[Response, int]]:

            log_prefix = f'[ VALIDATION | {request.remote_addr} {request.method} > {request.url_rule} ]'

            # skip decorator if unexpected HTTP method of request
            if method != request.method:
                logger.debug(f'{log_prefix} skip {request.method} method, validator expect {method}')
                return f(*args, **kw)

            # validate request data
            if request_schema:
                validation_schema = request_schema()

                if not request.is_json:
                    logger.warning(f'{log_prefix} < invalid content-type header')
                    return create_response(status=ServiceHTTPCodes.BAD_REQUEST.value,
                                           message='invalid content-type header')
                try:
                    validation_schema.load(request.json)
                    logger.debug(f'{log_prefix} request body validated')
                except exceptions.BadRequest:
                    logger.warning(f'{log_prefix} < invalid requests body')
                    return create_response(status=ServiceHTTPCodes.BAD_REQUEST.value,
                                           message='invalid requests body')
                except ValidationError as err:
                    logger.warning(f'{log_prefix} < validation error')
                    return create_response(status=ServiceHTTPCodes.UNPROCESSABLE_ENTITY.value,
                                           message='validation error',
                                           data=dict(err.messages))
            # call HTTP handler
            try:
                result = f(*args, **kw)
                if isinstance(result, Response):
                    result = create_response(data=result.json,
                                             status=result.status_code)
                response, status = result if isinstance(result, tuple) else create_response(result)
            except Exception as err:  # noqa
                logger.error(f'{log_prefix} < {err}')
                response, status = create_response(status=ServiceHTTPCodes.SERVER_ERROR.value,
                                                   message='unhandled error',
                                                   data=dict(
                                                       error=type(err).__name__,
                                                       traceback=str(traceback.format_exc())
                                                   ))
                return response, status

            # validate response data
            if response_schema:
                validation_schema = response_schema()

                if response.status_code != ServiceHTTPCodes.OK.value:
                    return response, status

                try:
                    data = validation_schema.load(response.json)
                    logger.debug(f'{log_prefix} response validated')
                    response = jsonify(validation_schema.dump(data))
                except (exceptions.BadRequest, ValidationError, TypeError) as err:
                    logger.error(f'{log_prefix} < {err}')
                    response, status = create_response(status=ServiceHTTPCodes.SERVER_ERROR.value,
                                                       message='incorrect response data',
                                                       data=dict(
                                                           error=type(err).__name__,
                                                           traceback=str(traceback.format_exc())
                                                       ))
                finally:
                    return response, status

            return response, status
        return wrapper

    return decorator



from datetime import datetime

from app import db


def format_text(text):
    return ' '.join(text.strip().split())


class Log(db.Model):
    __tablename__ = 'logs'
    id = db.Column(db.Integer(), primary_key=True)
    route = db.Column(db.String(100))
    method = db.Column(db.String(10))
    request = db.Column(db.Text)
    response = db.Column(db.Text)
    ts = db.Column(db.DateTime(), default=datetime.utcnow)
    duration = db.Column(db.Integer())
    status_code = db.Column(db.String)
    message = db.Column(db.String)

    def __repr__(self):
        return """<log object>
        id: {}
        route: {}
        method: {}
        request: {}
        response: {}
        ts: {}
        duration: {}
        status_code: {}
        message: {}""".format(self.id, self.route, self.method, self.request, self.response, self.ts,
                              self.duration, self.status_code, self.message)

    @staticmethod
    def init(request):
        log = Log()
        log.route = request.path
        log.method = request.method
        log.request = format_text(str(request.json))
        db.session.add(log)
        db.session.commit()
        return log.id

    @staticmethod
    def complete(log_id, response):
        log = Log.query.filter_by(id=log_id).first()
        if log:
            log.response = format_text(str(response.json))
            log.duration = (datetime.utcnow() - log.ts).total_seconds()
            log.status_code = response.status
            log.message = response.json['message']
            db.session.add(log)
            db.session.commit()
            return True
        return False

    def delete(self):
        log = Log.query.filter_by(id=self.id).first()
        db.session.delete(log)
        db.session.commit()



from unittest import TestCase

import requests
from flask import Response


class AbstractApp:
    client = None
    headers = {'Content-type': 'application/json'}

    def post(self, url, data=dict()):
        return self.client.post(url, json=data, headers=self.headers)

    def get(self, url, data=dict()):
        return self.client.get(url, json=data, headers=self.headers)

    def register(self, username, email, password, confirm):
        return self.post('/register', data=dict(username=username, email=email, password=password, confirm=confirm))

    def login(self, username, password):
        return self.post('/login', data=dict(username=username, password=password))

    def logout(self):
        return self.post('/logout')

    def delete(self, user_id, password):
        return self.post('/user/delete', data=dict(user_id=user_id, password=password))


class RemoteApp(AbstractApp):
    client = requests
    base_url = 'http://10.9.14.101:8000'

    def post(self, url, data=dict()):
        res = super().post(self.base_url + url, data)
        return self.convert_to_flask_response(res)

    def get(self, url, data=dict()):
        res = super().get(self.base_url + url, data)
        return self.convert_to_flask_response(res)

    # TODO transfer headers
    @staticmethod
    def convert_to_flask_response(response):
        res = Response(
            response=response.text,
            status=response.status_code,
            mimetype=response.headers['content-type']
        )

        return res


class UnitTest(TestCase, AbstractApp):
    app = None
    db = None

    def setUp(self):
        TestCase.setUp(self)
        try:
            print('SQLALCHEMY_DATABASE_URI: %s' % self.app.config['SQLALCHEMY_DATABASE_URI'])
        except KeyError:
            pass

        self.assertFalse(self.app is None)
        self.assertFalse(self.client is None)
        self.assertTrue(self.app.config['TESTING'])

        self.app_context = self.app.app_context()
        self.app_context.push()
        self.db.create_all()
        super().setUp()

    def tearDown(self):
        self.db.session.remove()
        self.db.drop_all()
        self.app_context.pop()
        TestCase.tearDown(self)


class BaseUnitTest(UnitTest):
    from app import create_app, db

    app = create_app('testing')
    client = app.test_client()
    db = db


class IntegrationTest(TestCase, RemoteApp):
    pass



import traceback

from flask import g, request

from .models import *
from . import logs


@logs.before_app_request
def before_request():
    try:
        log_id = Log.init(request)
        g.log_id = log_id
    except BaseException:
        print("Something went wrong during log initializing. See traceback: " + str(traceback.format_exc()))
        db.session.rollback()


@logs.after_app_request
def after_request(response):
    try:
        if g.log_id is not None:
            Log.complete(g.log_id, response)
        else:
            print('log_id is None, cannot complete logging process')
    except BaseException:
        print("Something went wrong during log completing. See traceback: " + str(traceback.format_exc()))
        db.session.rollback()
    return response



from typing import Tuple

import requests
from flask import jsonify
from flask.wrappers import Response


def safe_get(json, key):
    if key in json:
        return json[key]
    else:
        return None


class RequestTool:
    def post(self, url, data=dict()):
        headers = {'Content-type': 'application/json'}
        return requests.post(url, json=data, headers=headers)

    # TODO every service should register self in registry service
    def register_me(self, service_name, port=5000):
        return self.post('http://registry/register', dict(service_name=service_name, port=port))


class ResponseTool:

    def __init__(self, default_headers=dict()):
        self.default_headers = default_headers

    def create(self, data: dict = None, status: int = 200, message: str = "", headers=dict()) -> \
            Tuple[Response, int]:
        """Wraps response in a consistent format throughout the API.

        Format inspired by https://medium.com/@shazow/how-i-design-json-api-responses-71900f00f2db
        Modifications included:
        - make success a boolean since there's only 2 values
        - make message a single string since we will only use one message per response

        IMPORTANT: data must be a dictionary where:
        - the key is the name of the type of data
        - the value is the data itself

        :param headers <dict> Dictionary of headers
        :param data <str> optional data
        :param status <int> optional status code, defaults to 200
        :param message <str> optional message
        :returns tuple of Flask Response and int
        """

        if type(data) is not dict and data is not None:
            raise TypeError("Data should be a dictionary 😞")

        # merge default headers with current
        headers = {**self.default_headers, **headers}
        res = {"success": 200 <= status < 300, "message": message, "result": data,
               "headers": headers}
        response = jsonify(res)
        response.status_code = status
        return response, status

    def copy(self, response, message=None):
        if response is not None:
            return self.create(data=response.json()["result"], status=response.status_code,
                               message=message if message else response.json()["message"])
        else:
            raise TypeError("Response should not be None")


req = RequestTool()
post = req.post
register_me = req.register_me

resp = ResponseTool()
create_response = resp.create
copy_response = resp.copy



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="Avatar-Utils",
    version="1.0.8",
    author="Algorithmics of Complex System",
    author_email="artem.sementsov@gmail.com",
    description="Common utils for services in ecosystem",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="",
    packages=['avatar_utils.core', 'avatar_utils.db', 'avatar_utils.logs', 'avatar_utils.tests',
              'avatar_utils.validation'],
    install_requires=[
        'flask>=1.1.1',
        'flask-sqlalchemy>=2.3.2',
        'requests>=2.23.0',
        'psycopg2-binary>=2.7.5',
        'apispec==3.3.0',
        'marshmallow==3.6.0',
        'flasgger==0.9.4'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
)
